#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Component execution classes.
"""

class Context:

	"""
	Execution context.
	
	:param str base_dir:
	   Base directory.
	:param str prof_dir:
	   Profile working directory.
	"""
	
	def __init__(self, base_dir, prof_dir, prof_name):
	
		self.__base_dir = base_dir
		self.__prof_dir = prof_dir
		self.__prof_name = prof_name
		
	def base_dir(self):
	
		"""
		Base directory.
		
		:rtype:
		   str
		"""
		
		return self.__base_dir
		
	def profile_dir(self):
	
		"""
		Profile working directory.
		
		:rtype:
		   str
		"""
		
		return self.__prof_dir
		
	def profile_name(self):
	
		"""
		Profile name.
		
		:rtype:
		   str
		"""
		
		return self.__prof_name
		
class Operation:

	"""
	Component operation.
	
	:param str name:
	   Target component name.
	:param Component comp:
	   Target component. May be ``None``.
	"""
	
	def __init__(self, name, comp):
	
		self.__name = name
		self.__comp = comp
		if self.__comp is None:
			self.__execute = self.__execute_empty
		else:
			self.__execute = self.__execute_default
			
	def __eq__(self, other):
	
		return type(self) == type(other) and self.__name == other.__name
		
	def __ne__(self, other):
	
		return type(self) != type(other) or self.__name != other.__name
		
	def __repr__(self):
	
		return "{}.{}({}, {})".format(
			self.__module__,
			type(self).__name__,
			repr(self.__name),
			repr(self.__comp)
		)
		
	def __execute_empty(self, context):
	
		pass
		
	def __execute_default(self, context):
	
		self.execute_impl(self.__comp, context)
		
	def name(self):
	
		"""
		Component name.
		
		:rtype:
		   str
		"""
		
		return self.__name
		
	def execute(self, context):
	
		"""
		Execute operation.
		
		:param Context context:
		   Current execution context.
		"""
		
		self.__execute(context)
		
class Insert(Operation):

	"""
	Insert operation.
	
	:param str name:
	   Component name.
	:param Component comp:
	   Component to be inserted.
	"""
	
	def __init__(self, name, comp):
	
		super().__init__(name, comp)
		
	def execute_impl(self, comp, context):
	
		"""
		Execute insert operation.
		
		:param Component comp:
		   Target component.
		:param Context context:
		   Current context.
		"""
		
		comp.insert(context)
		
class Delete(Operation):

	"""
	Delete operation.
	
	:param str name:
	   Component name.
	:param Component comp:
	   Component to be deleted.
	"""
	
	def __init__(self, name, comp):
	
		super().__init__(name, comp)
		
	def execute_impl(self, comp, context):
	
		"""
		Execute delete operation.
		
		:param Component comp:
		   Target component.
		:param Context context:
		   Current context.
		"""
		
		comp.delete(context)




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Package containing engine classes.
"""

import configparser
import os

class ProfileManager:

	"""
	Manager for profiles located in a directory.
	
	:param str base_dir:
	   Directory path of this manager. Must be absolute.
	:param str work_dir:
	   Working directory. From base_dir if it is relative.
	:param dict profiles:
	   Dictionary containing profile type and instance entries.
	"""
	
	def __init__(self, base_dir, work_dir, profiles):
	
		if not os.path.isabs(base_dir):
			msg = "{} is not absolute".format(base_dir)
			raise Exception(msg)
		self.__base_dir = base_dir
		
		if os.path.isabs(work_dir):
			self.__work_dir = work_dir
		else:
			self.__work_dir = os.path.join(self.__base_dir, work_dir)
			
		self.__profiles = profiles or {}
		
	def __profile_stub(self, prof_name, prof_type, prof_dir):
	
		try:
			prof = self.__profiles[prof_type]
			return ProfileStub(self.__base_dir, prof_name, prof, prof_dir)
		except KeyError:
			msg = "Profile of type '{}' does not exist"
			raise Exception(msg.format(prof_type))
			
	def __profile_dir(self, prof_name):
	
		return os.path.join(self.__work_dir, prof_name)
		
	def __profile_ini_path(self, prof_dir):
	
		return os.path.join(prof_dir, "profile.ini")
		
	def load(self, prof_name):
	
		"""
		Load the profile with the given name.
		
		:param str prof_name:
		   Profile name.
		:rtype:
		   ProfileStub
		:return:
		   An stub to loaded profile.
		"""
		
		prof_dir = self.__profile_dir(prof_name)
		prof_ini_path = self.__profile_ini_path(prof_dir)
		if not os.path.exists(prof_ini_path):
			msg = "Profile '{}' does not exist"
			raise Exception(msg.format(prof_name))
			
		# Load profile
		prof_ini_file = open(prof_ini_path, "r")
		prof_ini = configparser.ConfigParser()
		prof_ini.read_file(prof_ini_file)
		prof_ini_file.close()
		
		# Prepare profile
		prof_type = prof_ini["profile"]["type"]
		prof_stub = self.__profile_stub(prof_name, prof_type, prof_dir)
		prof_stub.prepare(prof_ini["properties"])
		
		return prof_stub
		
	def store(self, prof_name, prof_type):
	
		"""
		Store a profile with the given name and type.
		
		:param str prof_name:
		   Profile name.
		:param str prof_type:
		   Profile type.
		"""
		
		prof_dir = self.__profile_dir(prof_name)
		prof_stub = self.__profile_stub(prof_name, prof_type, prof_dir)
		if not os.path.exists(prof_dir):
			os.makedirs(prof_dir)
		prof_ini_path = self.__profile_ini_path(prof_dir)
		
		# Load previous properties
		if os.path.exists(prof_ini_path):
			prof_ini_file = open(prof_ini_path, "r")
			prof_ini = configparser.ConfigParser()
			prof_ini.read_file(prof_ini_file)
			prof_ini_file.close()
			prev_props = prof_ini["properties"]
		else:
			prev_props = {}
			
		# Prepare and store profile
		prof_ini = configparser.ConfigParser()
		prof_ini["profile"] = {}
		prof_ini["profile"]["type"] = prof_type
		prof_ini["properties"] = prof_stub.prepare(prev_props)
		prof_ini_file = open(prof_ini_path, "w")
		prof_ini.write(prof_ini_file)
		prof_ini_file.close()
		
class ProfileStub:

	"""
	Stub of a :class:`Profile` with a :class:`ComponentStubRegistry`.
	
	:param str prof_name:
	   Profile name.
	:param Profile prof:
	   Profile instance.
	:param str prof_dir:
	   Profile working directory.
	"""
	
	def __init__(self, base_dir, prof_name, prof, prof_dir):
	
		self.__base_dir = base_dir
		self.__prof_name = prof_name
		self.__prof = prof
		self.__prof_dir = prof_dir
		self.__comp_stub_reg = ComponentStubRegistry(self.__prof)
		
	def __component_list(self, comp_stub, target_list):
	
		if comp_stub not in target_list:
			target_list.append(comp_stub)
		for dep_stub in comp_stub.dependencies():
			self.__component_list(dep_stub, target_list)
			
	def context(self):
	
		"""
		Create an exectution context.
		
		:rtype:
		   execution.Context
		:return:
		   The created execution context.
		"""
		
		return execution.Context(self.__base_dir, self.__prof_dir,
				self.__prof_name)
		
	def prepare(self, props):
	
		"""
		Prepare stubbed profile and return its properties.
		
		:param dict props:
		   Dictionary with previous properties.
		:rtype:
		   dict
		:return:
		   Properties dictionary.
		"""
		
		return self.__prof.prepare(props)
		
	def component_list(self):
	
		"""
		List of :class:`ComponentStub` instances for this profile.
		
		:rtype:
		   list
		"""
		
		comp_list = []
		self.__component_list(self.component(None), comp_list)
		return sorted(comp_list)
		
	def component(self, comp_name):
	
		"""
		Component with the specified name.
		
		:param str comp_name:
		   Specified component name.
		:rtype:
		   ComponentStub
		:return:
		   The component stub with the specified name.
		"""
		
		return self.__comp_stub_reg.get(comp_name)
		
class ComponentStub:

	"""
	Stub of a :class:`Component`.
	
	:param ComponentStubRegistry comp_stub_reg:
	   Component stub registry.
	:param str comp_name:
	   Component name.
	:param Profile prof:
	   Owner profile.
	"""
	
	def __init__(self, comp_stub_reg, comp_name, prof):
	
		self.__comp_stub_reg = comp_stub_reg
		self.__comp_name = comp_name
		self.__comp_inst = None
		self.__comp = self.__comp_init
		self.__prof = prof
		self.__type_name = self.__type_name_init
		self.__available = self.__available_init
		
	def __eq__(self, other):
	
		return self.name() == other.name()
		
	def __ne__(self, other):
	
		return self.name() != other.name()
		
	def __lt__(self, other):
	
		if self.name() is None or other.name() is None:
			return None
		return self.name() < other.name()
		
	def __comp_init(self):
	
		self.__comp_inst = self.__prof.component(self.__comp_name)
		return self.__comp_inst
		
	def __comp_default(self):
	
		return self.__comp_inst
		
	def __type_name_init(self):
	
		comp = self.__comp()
		if comp is None:
			self.__type_name = self.__type_name_empty
		else:
			self.__type_name = self.__type_name_default
		return self.__type_name()
		
	def __type_name_empty(self):
	
		return None
		
	def __type_name_default(self):
	
		return "{}.{}".format(
			self.__comp_inst.__module__,
			self.__comp_inst.__class__.__name__
		)
		
	def __available_init(self, context):
	
		comp = self.__comp()
		if comp is None:
			self.__available = self.__available_empty
		else:
			self.__available = self.__available_default
		return self.__available(context)
		
	def __available_empty(self, context):
	
		return None
		
	def __available_default(self, context):
	
		return self.__comp_inst.available(context)
		
	def __dependents(self, comp_stub, dep_list):
	
		for dep_stub in comp_stub.dependencies():
			if dep_stub == self:
				dep_list.append(comp_stub)
			else:
				self.__dependents(dep_stub, dep_list)
				
	def name(self):
	
		"""
		Component name.
		
		:rtype:
		   str
		"""
		
		return self.__comp_name
		
	def type_name(self):
	
		"""
		Fully qualified name of component class or ``None``.
		
		:rtype:
		   str
		"""
		
		return self.__type_name()
		
	def dependencies(self):
	
		"""
		Yield :class:`ComponentStub` of dependencies of this component.
		"""
		
		try:
			for dep_name in self.__prof.dependencies(self.name()):
				yield self.__comp_stub_reg.get(dep_name)
		except TypeError:
			yield from ()
			
	def available(self, context):
	
		"""
		Return component availability.
		
		:rtype:
		   bool
		:return:
		   Component availability.
		"""
		
		return self.__available(context)
		
	def insert(self, context, plan):
	
		"""
		Include an insert operation to the given plan.
		
		:param execution.Context context:
		   Current execution context.
		:param list plan:
		   List of :class:`execution.Operation` instances.
		"""
		
		op = execution.Insert(self.__comp_name, self.__comp())
		if op not in plan and self.available(context) != True:
			for dep_stub in self.dependencies():
				dep_stub.insert(context, plan)
			plan.append(op)
			
	def delete(self, context, plan):
	
		"""
		Include a delete operation to the given plan.
		
		:param execution.Context context:
		   Current execution context.
		:param list plan:
		   List of :class:`execution.Operation` instances.
		"""
		
		op = execution.Delete(self.__comp_name, self.__comp())
		dep_list = []
		self.__dependents(self.__comp_stub_reg.get(None), dep_list)
		if op not in plan and self.available(context) == True:
			for dep_stub in dep_list:
				dep_stub.delete(context, plan)
			plan.append(op)
			
class ComponentStubRegistry:

	"""
	Registry of :class:`ComponentStub`.
	
	:param Profile prof:
	   Involved profile.
	"""
	
	def __init__(self, prof):
	
		self.__prof = prof
		self.__cache = {}
		
	def get(self, comp_name):
	
		"""
		Get the specified component.
		
		:param str comp_name:
		   Specified component name.
		:rtype:
		   ComponentStub
		:return:
		   Specified component stub.
		"""
		
		try:
			return self.__cache[comp_name]
		except KeyError:
			comp_stub = ComponentStub(self, comp_name, self.__prof)
			self.__cache[comp_name] = comp_stub
			return comp_stub




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import resort.engine
import resort.engine.execution

import argparse
import colorama
import io
import os
import sys

#
# Entry point
#
def setup(
	work_dir=".resort",
	profiles=None
):

	"""
	Configures a project to be managed by *resort*.
	
	:param str work_dir:
	   Path of directory were profiles will be created.
	:param dict profiles:
	   Dictionary containing profile type and instance entries.
	"""
	
	# Initialize colorama
	colorama.init()
	
	# Program name
	prog_name = sys.argv[0]
	
	# Profile manager
	prof_mgr = engine.ProfileManager(os.getcwd(), work_dir, profiles)
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"profile",
		metavar="profile",
		type=str,
		nargs=1,
		help="Profile name"
	)
	parser.add_argument(
		"command",
		metavar="command",
		type=argparse_command,
		nargs=1,
		help="Command name"
	)
	parser.add_argument(
		"arguments",
		metavar="args",
		nargs=argparse.REMAINDER,
		help="Command arguments"
	)
	args = parser.parse_args(sys.argv[1:])
	
	# Execute command
	cmd_name, cmd_fn = args.command[0]
	cmd_prog_name = "{} profile {}".format(prog_name, cmd_name)
	prof_name = args.profile[0]
	prog_args = args.arguments
	cmd_fn(cmd_prog_name, prof_mgr, prof_name, prog_args)
	
#
# Argument parser for command
#
def argparse_command(value):

	try:
		cmd_prefix = "command_"
		return (value, globals()["{}{}".format(cmd_prefix, value)])
	except:
		msg = io.StringIO()
		msg.write("unrecognized command '")
		msg.write(value)
		msg.write("'\n\nAvailable commands:\n")
		cmds = [
			(g_name[len(cmd_prefix):], globals()[g_name])
			for g_name in sorted(globals())
			if g_name.startswith(cmd_prefix)
		]
		desc_pos = 0
		for cmd_name, cmd in cmds:
			desc_pos = max(desc_pos, len(cmd_name))
		for cmd_name, cmd in cmds:
			msg.write(cmd_name.ljust(desc_pos))
			msg.write("  ")
			msg.write(cmd.__doc__.strip())
			msg.write("\n")
		msg.seek(0)
		raise argparse.ArgumentTypeError(msg.read())
		
#
# Argument parser for status depth
#
def argparse_status_depth(value):

	try:
		if int(value) > 0:
			return int(value)
	except:
		pass
	raise argparse.ArgumentTypeError("Bad tree depth {}".format(value))
	
#
# Component existence check
#
def component_exists(prof_stub, comp_stub):

	if comp_stub not in prof_stub.component_list():
		msg = "Component \"{}\" does not exist"
		raise Exception(msg.format(comp_stub.name()))
		
#		
# Component status printer
#
def print_component_status(out, context, comp_stub, last, depth, indent,
		show_type, tree_depth):

	for indent_last in indent:
		if indent_last:
			out.write(" ")
		else:
			out.write("\u2502")
		out.write("   ")
	if last:
		out.write("\u2514")
	else:
		out.write("\u251c")
	out.write("\u2500 ")
	
	avail = comp_stub.available(context)
	if avail is True:
		out.write(colorama.Style.BRIGHT)
	elif avail is False:
		out.write(colorama.Style.DIM)
	out.write(comp_stub.name())
	
	type_name = comp_stub.type_name()
	if show_type and type_name is not None:
		out.write(" (")
		out.write(type_name)
		out.write(")")
	out.write(colorama.Style.RESET_ALL)
	out.write("\n")
	
	deps = list(comp_stub.dependencies())
	new_depth = depth + 1
	new_indent = []
	new_indent.extend(indent)
	new_indent.append(last)
	for i, dep_stub in enumerate(deps):
		last = i == len(deps) - 1
		if tree_depth is None or depth < tree_depth:
			print_component_status(out, context, dep_stub, last, new_depth,
					new_indent, show_type, tree_depth)
					
#
# Command init
#
def command_init(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Initialize a profile.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"type",
		metavar="type",
		type=str,
		nargs=1,
		help="profile type"
	)
	args = parser.parse_args(prog_args)
	
	# Profile store
	prof_type = args.type[0]
	prof_mgr.store(prof_name, prof_type)
	
#
# Command list
#
def command_list(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Print the list of components.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	args = parser.parse_args(prog_args)
	
	# Profile load
	prof_stub = prof_mgr.load(prof_name)
	
	# Print component list
	out = io.StringIO()
	for comp_stub in prof_stub.component_list():
		if comp_stub.name() is not None:
			out.write(comp_stub.name())
			out.write("\n")
	out.seek(0)
	sys.stdout.write(out.read())
	
#
# Command status
#
def command_status(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Show status of component tree.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"-t",
		"--type",
		required=False,
		action="store_true",
		default=False,
		dest="show_type",
		help="show component qualified class name"
	)
	parser.add_argument(
		"-d",
		"--depth",
		metavar="tree_depth",
		required=False,
		type=argparse_status_depth,
		default=None,
		dest="tree_depth",
		help="integer value of dependency tree depth"
	)
	parser.add_argument(
		"components",
		metavar="comps",
		nargs=argparse.REMAINDER,
		help="system components"
	)
	args = parser.parse_args(prog_args)
	
	# Profile load
	prof_stub = prof_mgr.load(prof_name)
	
	# Collect component stubs
	comp_stubs = []
	if len(args.components) == 0:
		raise Exception("Empty component list")
	for comp_name in args.components:
		comp_stub = prof_stub.component(comp_name)
		component_exists(prof_stub, comp_stub)
		comp_stubs.append(comp_stub)
		
	# Print status
	show_type = args.show_type
	tree_depth = args.tree_depth
	context = prof_stub.context()
	out = io.StringIO()
	for i, comp_stub in enumerate(comp_stubs):
		last = i == len(comp_stubs) - 1
		print_component_status(out, context, comp_stub, last, 1, [], show_type,
				tree_depth)
	out.seek(0)
	sys.stdout.write(out.read())
	
#
# Command insert
#
def command_insert(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Insert components.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"components",
		metavar="comps",
		nargs=argparse.REMAINDER,
		help="system components"
	)
	args = parser.parse_args(prog_args)
	
	# Profile load
	prof_stub = prof_mgr.load(prof_name)
	
	# Collect component stubs
	comp_stubs = []
	if len(args.components) == 0:
		raise Exception("Empty component list")
	for comp_name in args.components:
		comp_stub = prof_stub.component(comp_name)
		component_exists(prof_stub, comp_stub)
		comp_stubs.append(comp_stub)
		
	# Create insert plan
	context = prof_stub.context()
	plan = []
	for comp_stub in comp_stubs:
		comp_stub.insert(context, plan)
		
	# Print insert plan
	for op in plan:
		op.execute(context)
		
#
# Command delete
#
def command_delete(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Delete components.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"components",
		metavar="comps",
		nargs=argparse.REMAINDER,
		help="system components"
	)
	args = parser.parse_args(prog_args)
	
	# Profile load
	prof_stub = prof_mgr.load(prof_name)
	
	# Collect component stubs
	comp_stubs = []
	if len(args.components) == 0:
		raise Exception("Empty component list")
	for comp_name in args.components:
		comp_stub = prof_stub.component(comp_name)
		component_exists(prof_stub, comp_stub)
		comp_stubs.append(comp_stub)
			
	# Create delete plan
	context = prof_stub.context()
	plan = []
	for comp_stub in comp_stubs:
		comp_stub.delete(context, plan)
		
	# Print delete plan
	for op in plan:
		op.execute(context)
		
#
# Command update
#
def command_update(prog_name, prof_mgr, prof_name, prog_args):

	"""
	Update components.
	"""
	
	# Retrieve arguments
	parser = argparse.ArgumentParser(
		prog=prog_name
	)
	parser.add_argument(
		"components",
		metavar="comps",
		nargs=argparse.REMAINDER,
		help="system components"
	)
	args = parser.parse_args(prog_args)
	
	# Profile load
	prof_stub = prof_mgr.load(prof_name)
	
	# Collect component stubs
	comp_stubs = []
	if len(args.components) == 0:
		raise Exception("Empty component list")
	for comp_name in args.components:
		comp_stub = prof_stub.component(comp_name)
		component_exists(prof_stub, comp_stub)
		comp_stubs.append(comp_stub)
		
	context = prof_stub.context()
	
	# Create delete plan
	plan = []
	for comp_stub in comp_stubs:
		comp_stub.delete(context, plan)
		
	# Execute delete plan
	for op in plan:
		op.execute(context)
		
	# Update component stub list
	for op in plan:
		comp_stub = prof_stub.component(op.name())
		if comp_stub not in comp_stubs:
			comp_stubs.append(comp_stub)
	
	# Create insert plan
	plan = []
	for comp_stub in comp_stubs:
		comp_stub.insert(context, plan)
		
	# Execute insert plan
	for op in plan:
		op.execute(context)

